// Copyright 2022 Harness Inc. All rights reserved.
// Use of this source code is governed by the Polyform Free Trial License
// that can be found in the LICENSE.md file for this repository.

package space

import (
	"context"
	"fmt"
	"strings"

	apiauth "github.com/harness/gitness/internal/api/auth"
	"github.com/harness/gitness/internal/auth"
	"github.com/harness/gitness/types"
	"github.com/harness/gitness/types/check"
	"github.com/harness/gitness/types/enum"
)

// UpdateInput is used for updating a space.
type UpdateInput struct {
	Description *string `json:"description"`
	IsPublic    *bool   `json:"is_public"`
}

func (in *UpdateInput) hasChanges(space *types.Space) bool {
	return (in.Description != nil && *in.Description != space.Description) ||
		(in.IsPublic != nil && *in.IsPublic != space.IsPublic)
}

// Update updates a space.
func (c *Controller) Update(ctx context.Context, session *auth.Session,
	spaceRef string, in *UpdateInput) (*types.Space, error) {
	space, err := c.spaceStore.FindByRef(ctx, spaceRef)
	if err != nil {
		return nil, err
	}

	if err = apiauth.CheckSpace(ctx, c.authorizer, session, space, enum.PermissionSpaceEdit, false); err != nil {
		return nil, err
	}

	if !in.hasChanges(space) {
		return space, nil
	}

	if err = sanitizeUpdateInput(in); err != nil {
		return nil, fmt.Errorf("failed to sanitize input: %w", err)
	}

	space, err = c.spaceStore.UpdateOptLock(ctx, space, func(space *types.Space) error {
		// update values only if provided
		if in.Description != nil {
			space.Description = *in.Description
		}
		if in.IsPublic != nil {
			space.IsPublic = *in.IsPublic
		}

		return nil
	})
	if err != nil {
		return nil, err
	}

	return space, nil
}

func sanitizeUpdateInput(in *UpdateInput) error {
	if in.Description != nil {
		*in.Description = strings.TrimSpace(*in.Description)
		if err := check.Description(*in.Description); err != nil {
			return err
		}
	}

	return nil
}
